@@ -2,8 +2,7 @@
 
 Written by Pierre Quentel quentel.pierre@wanadoo.fr
 
-Published under the GPL licence - no warranty, used at user's risk etc. See
-http://www.gnu.org/copyleft/gpl.html and the file LICENCE.txt
+Published under the BSD licence. See the file LICENCE.txt
 
 Request handler built upon SimpleHTTPRequestHandler, except for the following
 extensions :
@@ -53,37 +52,25 @@ sessionObject=Session()
 A maximum of 1000 simultaneous sessions is supported
 """
 
-import __builtin__
-import sys, os, string, cStringIO, traceback, base64, time, gettext, copy, imp
-import Cookie, urlparse, mimetypes, BaseHTTPServer, cgi, urllib, threading
+import sys, os, string, cStringIO, traceback, base64, time, gettext, copy
+import Cookie, urlparse, mimetypes, BaseHTTPServer, cgi, urllib
+import logging.handlers
 
-import Template, URLResolution, k_utils, debugger.k_debugger, k_session
-import k_config
-import modify_request
-from InternationalRequestHandler import InternationalRequestHandler
+# Karrigell-sepcific modules
+from core import k_config
+import Template, URLResolution, k_utils, k_session, k_script, modify_request
+import debugger.k_debugger
+from k_encodings import k_encoding_charsets, k_encoding
+from k_stringio import KStringIO as StringIO
+from LocalizedRequestHandler import LocalizedRequestHandler
 
-__version__ = "2.2.1"
+__version__ = "2.2.5"
 
 os.chdir(k_config.serverDir)
 
 class ReloadError(Exception):
     pass
 
-class HTTP_REDIRECTION(Exception):
-    pass
-
-class AUTH_ABORT(Exception):
-    pass
-
-class HTTP_ERROR(Exception):
-
-    def __init__(self,code,message):
-        self.code=code
-        self.message=message
-
-    def __str__(self):
-        return self.code+self.message
-
 # update mime types
 mimetypes.init()
 mimetypes.types_map.update({
@@ -96,19 +83,36 @@ mimetypes.types_map.update({
     })
 mimetypes.types_map.update(k_config.extensions_map)
 
-# adds current directory and databases directory to sys.path,
 # for imports
-paths=[os.getcwd(), os.path.join(k_config.serverDir,"debugger")]
+if not os.getcwd() in sys.path:
+    sys.path.append(os.getcwd())
 
-for p in paths:
-    if not p in sys.path:
-        sys.path.append(p)
+# set up logging
+if k_config.loggingFile:
+    _logdir = os.path.dirname(k_config.loggingFile)
+    if not os.path.exists(_logdir):
+        os.makedirs(_logdir)
+    pars = eval(k_config.loggingParameters) # ugly way
+    loghandler = logging.handlers.RotatingFileHandler(k_config.loggingFile, *pars)
+    loghandler.setLevel(logging.INFO)
+    logging.getLogger().setLevel(logging.INFO)
+    logging.getLogger().addHandler(loghandler)
 
-class KarrigellRequestHandler(InternationalRequestHandler):
+def content_type():
+    "return content type for html texts"
+    if not k_config.outputEncoding:
+        ct="text/html"
+    else:
+        ct="text/html; charset="+k_encoding_charsets.encoding2mime_map.get(
+            k_config.outputEncoding, k_config.outputEncoding)
+    return ct
+
+class KarrigellRequestHandler(LocalizedRequestHandler):
 
     server_version = "Karrigell/" + __version__
     cachemanaged   = 1      # httpd optimization
     imported = {}
+    accepted_languages = []
 
     if k_config.language:
         accepted_languages.append(k_config.language)
@@ -125,15 +129,47 @@ class KarrigellRequestHandler(InternationalRequestHandler):
     def prepare_env(self):
         """Prepare environment for dynamic scripts"""
 
-        # initialize the cookie
+        # initialize the cookie objects
+        # COOKIE = cookies received from the browser
         if self.HEADERS.has_key("cookie"):
-            self.SET_COOKIE=Cookie.SimpleCookie(self.HEADERS["cookie"])
+            self.COOKIE=Cookie.SimpleCookie(self.HEADERS["cookie"])
+        else:
+            self.COOKIE=Cookie.SimpleCookie()
+        # SET_COOKIE = cookies sent back to the browser            
+        self.SET_COOKIE=Cookie.SimpleCookie()
 
         # if there is a Query String, decode it in a QUERY dictionary
         # handle the query string, if any
         self.QUERY = cgi.parse_qs(self.qs,1)
 
-        # if the method was POST the request body is in self.body
+        # all unicode management written by Radovan Garabik 
+        # and Laurent Pointal
+        if k_config.encodeFormData:
+            # try encoding the query string
+            qs_encoding = k_encoding.try_encoding(urllib.unquote(self.qs), 
+               ['ascii', k_config.outputEncoding, 'utf-8', 'iso8859_1_ncc', 
+               'cp1252', 'macroman'])
+
+            # if qs_encoding is None, encoding could not have been determined
+            # most likely it is gibberish, or a malicious user entered invalid 
+            # sequence intentionally
+            # in any case, we just have no idea what to do with the QUERY, so 
+            # we might as well discard it....
+            if qs_encoding is None:
+                self.QUERY = {}
+        
+            # now fix the query, convert it to unicode using our guessed 
+            # encoding
+            for key, val in self.QUERY.items():
+                for i, v in enumerate(val): 
+                    try:
+                        # if it is just an ascii string, leave it in peace
+                        unicode(v, 'ascii')
+                        continue
+                    except UnicodeDecodeError:
+                        # modify val list in-place
+                        val[i] = unicode(v, qs_encoding)
+
         if self.command == 'POST':
             for key in self.body.keys():
                 self.QUERY[key] = self.body[key]
@@ -142,19 +178,37 @@ class KarrigellRequestHandler(InternationalRequestHandler):
                 for i,elt in enumerate(self.QUERY[key]):
                     if not elt.filename:
                         # convert FieldStorage to string if not file upload
-                        self.QUERY[key][i] = elt.value 
+                        s = elt.value
+                        # elt.value SHOULD be in the same encoding as 
+                        # outputEncoding, but unfortunately there are some 
+                        # broken browsers so we have to do the test...
+                        # for backward compatibility, if outputEncoding=='',
+                        # don't use unicode at all
+                        if k_config.encodeFormData:
+                            s_encoding = k_encoding.try_encoding(s, 
+                                ['ascii', k_config.outputEncoding, 'utf-8', 
+                                'iso8859_1_ncc', 'cp1252', 'macroman'])
+                        else:
+                            s_encoding = 'ascii'
+                        if s_encoding == 'ascii':
+                            self.QUERY[key][i] = s
+                        elif s_encoding:
+                            self.QUERY[key][i] = unicode(s, s_encoding)
+                        else: # gibberish
+                            self.QUERY[key][i] = '' # or would it be better 
+                                                    # to delete it?
 
+        self.QUERY = modify_request.modify_query(self.QUERY)
         self.QUERY=k_utils.applyQueryConvention(self.QUERY)
         
-        self.ctype='text/html'  # default content-type
+        self.ctype=content_type()  # default content-type
 
         # replace standard output by a StringIO
         # the "print" statements in scripts will write to this StringIO
-        self.outputStream=cStringIO.StringIO()
+        self.outputStream=StringIO()
 
         # select language
         self.get_language()
-        #self.translateClassAttributes()
 
         # read an Authorization header if any
         # and decode it in AUTH_USER and AUTH_PASSWORD
@@ -172,11 +226,9 @@ class KarrigellRequestHandler(InternationalRequestHandler):
         self.nameSpace.update({
             "RESPONSE":self.RESPONSE,"HEADERS":self.HEADERS,
             "AUTH_USER":self.AUTH_USER,"AUTH_PASSWORD":self.AUTH_PASSWORD,
-            "QUERY":self.QUERY,"SET_COOKIE":self.SET_COOKIE,
+            "QUERY":self.QUERY,"COOKIE":self.COOKIE,
+            "SET_COOKIE":self.SET_COOKIE,
             "ACCEPTED_LANGUAGES":self.accepted_languages,
-            "HTTP_REDIRECTION":HTTP_REDIRECTION,
-            "HTTP_ERROR":HTTP_ERROR,
-            "AUTH_ABORT":AUTH_ABORT,
             "SERVER_DIR":k_config.serverDir,
             "Session":self.Session,"Authentication":self.Authentication,
             "os":os,"Cookie":Cookie,"string":string})
@@ -187,11 +239,70 @@ class KarrigellRequestHandler(InternationalRequestHandler):
         to prevent crashing the server"""
         try:
             self.try_handle_data()
-        except AUTH_ABORT:
+        except k_script.AUTH_ABORT:
             pass
         except:
             traceback.print_exc(file=sys.stderr)
 
+    def log_date_time_string(self):
+        """Return the current time formatted for logging."""
+        now = time.time()
+        year, month, day, hh, mm, ss, x, y, z = time.localtime(now)
+        tzh = -(time.timezone/3600)
+        tzm = (time.timezone%3600)/60
+        s = "%02d/%3s/%04d:%02d:%02d:%02d %+03d%02d" % (
+                day, self.monthname[month], year, hh, mm, ss, tzh, tzm)
+        return s
+
+    def log_error(self, format, *args):
+        """Log an error.
+
+        This is called when a request cannot be fulfilled.  By
+        default it passes the message on to log_message().
+
+        Arguments are the same as for log_message().
+        
+        Put the error message only to stderr.
+
+        """
+        if not k_config.silent:
+            msg = ("%s - - [%s] %s" %
+                             (self.address_string(),
+                              self.log_date_time_string(),
+                              format%args))
+
+            sys.stderr.write(msg)
+            sys.stderr.write('\n')
+
+        
+
+    def log_message(self, format, *args):
+        """Log an arbitrary message.
+
+        This is used by all other logging functions.  Override
+        it if you have specific logging wishes.
+
+        The first argument, FORMAT, is a format string for the
+        message to be logged.  If the format string contains
+        any % escapes requiring parameters, they should be
+        specified as subsequent arguments (it's just like
+        printf!).
+
+        The client host and current date/time are prefixed to
+        every message.
+
+        """
+
+        msg = ("%s - - [%s] %s" %
+                         (self.address_string(),
+                          self.log_date_time_string(),
+                          format%args))
+        if not k_config.silent:
+            BaseHTTPServer.BaseHTTPRequestHandler.log_message(self, format, *args)
+
+        if k_config.loggingFile:
+            logging.info(msg)
+
     def try_handle_data(self):
         """Handle a GET or POST request"""
         sys.stdout = k_utils.Stdout()
@@ -201,22 +312,29 @@ class KarrigellRequestHandler(InternationalRequestHandler):
         self.path = modify_request.modify_path(self.path)
         self.HEADERS = modify_request.modify_headers(self.HEADERS)
 
-        self.RESPONSE=k_utils.CI_dict({'Content-Type':"text/html"})  # default value
+        self.RESPONSE=k_utils.CI_dict({'Content-Type':content_type()})  # default value
         parsed_path = urlparse.urlparse(self.path)
         self.qs = parsed_path[4]
-        self.path_without_qs = parsed_path[2]
         
+        path_without_qs = parsed_path[2]
+
+        self.path_without_qs = path_without_qs
+
         # virtual hosts
         self.host = self.HEADERS.get('host',0)
+        if self.host !=0 and ':' in self.host:
+            self.host = self.host.split(':')[0]
         if not k_config.virtual_hosts.has_key(self.host):
             self.host = 0
         for k in k_config.virtual_hosts[self.host].keys():
             setattr(k_config,k,k_config.virtual_hosts[self.host][k])
 
         fileName=URLResolution.translate_path(self.path_without_qs)
+        subpath = URLResolution.getScriptElements(self.path_without_qs)
+        self.COOKIE = Cookie.SimpleCookie()
         self.SET_COOKIE=Cookie.SimpleCookie()
         self.nameSpace={"REQUEST_HANDLER":self}
-        self.outputStream = cStringIO.StringIO()
+        self.outputStream = StringIO()
 
         # save current directory (it may be modified inside scripts)
         saveDir=os.getcwd()
@@ -253,13 +371,14 @@ class KarrigellRequestHandler(InternationalRequestHandler):
                     self.send_error(403,"Directory listing not allowed")
                     return
                 try:
-                    f = cStringIO.StringIO(
+                    f = StringIO()
+                    f.write(
                         Template.list_directory(self.path_without_qs,fileName))
-                    self.ctype = 'text/html'
+                    self.ctype = content_type()
                 except:
-                    f = cStringIO.StringIO()
+                    f = StringIO()
                     traceback.print_exc(file = f)
-                    self.ctype = 'text/plain'
+                    self.ctype = content_type()
                 self.outputStream = f
                 self.outputStream.read()    # so that tell() returns the length
                 self.karrigellSendResponse(200,"Ok")
@@ -275,8 +394,7 @@ class KarrigellRequestHandler(InternationalRequestHandler):
 
         # if file is a ks script with no method, redirect to method 'index'
         if os.path.isfile(fileName) and fileName.lower().endswith('.ks'):
-            function=URLResolution.getKsFunction(self.path)
-            if not function:
+            if not len(subpath):
                 if not self.path_without_qs.endswith('/'):
                     self.path=self.path_without_qs+'/index'
                 else:
@@ -298,7 +416,18 @@ class KarrigellRequestHandler(InternationalRequestHandler):
                 "the files with this extension")
             return
             
+        # check if path authorized
+        # hidden paths are regular expression patterns specified in 
+        # [Directories] hidden_extensions in the configuration file
+        for hidden_path in k_config.hide_paths:
+            if hidden_path.match(self.path_without_qs):
+                self.send_error(403,"You are not allowed to see " \
+                "the files with the pattern %s" %hidden_path.pattern)
+                return
+
         self.ctype, self.encoding = mimetypes.guess_type(fileName)
+        if self.ctype=='text/html':
+            self.ctype = content_type()
         self.RESPONSE['Content-Type']=self.ctype
 
         os.chdir(os.path.dirname(fileName))
@@ -308,15 +437,15 @@ class KarrigellRequestHandler(InternationalRequestHandler):
         if self.cache(fileName):
             script = self.loadedScripts[fileName][1]
         # process the script according to its extension
-        elif extension.lower() in Template.handled_extensions:
+        elif extension.lower() in k_config.handled_extensions:
             # Python, pih or hip script : prepare environment and namespace
             try:
                 script=Template.getScript(fileName)
                 self.loadedScripts[fileName]=(os.path.getmtime(fileName),
                     script)
-            except Template.ParseError,error:
+            except k_script.ParseError,error:
                 # parse error : show traceback and file content
-                tb=cStringIO.StringIO()
+                tb=StringIO()
                 traceback.print_exc(file=tb)
                 error_info = [self.path,fileName,tb.getvalue(),
                     error.errorLine]
@@ -326,12 +455,11 @@ class KarrigellRequestHandler(InternationalRequestHandler):
                     err_script = Template.getScript(errorFileName)
                     out = err_script.render({'error_info':error_info}).value
                 except:
-                    out = cStringIO.StringIO()
+                    out = StringIO()
                     traceback.print_exc(file=out)
                     out = out.getvalue()
-                self.outputStream = cStringIO.StringIO()
+                self.outputStream = StringIO()
                 self.outputStream.write(out)
-                k_utils.trace(out)
                 self.karrigellSendResponse(200,'Ok')
                 return
         
@@ -341,6 +469,7 @@ class KarrigellRequestHandler(InternationalRequestHandler):
             script.url=self.path_without_qs
             script.path=self.path
             script.parent=None
+            script.subpath = subpath
             # execution
             self.execute(script)
         elif k_utils.pathInDirs(fileName,k_config.protectedDirs)[0]:
@@ -349,7 +478,7 @@ class KarrigellRequestHandler(InternationalRequestHandler):
             self.nameSpace['fileName'] = fileName
             depth = k_utils.pathInDirs(fileName,k_config.protectedDirs)[1]
             script = Template.getScript(
-                os.path.join(k_config.serverDir,'dump.py'))
+                os.path.join(k_config.serverDir,'core','dump.py'))
             script.code = 'Include("'+'../'*depth + 'AuthentScript.py");' \
                 + script.code
             self.execute(script)
@@ -431,30 +560,40 @@ class KarrigellRequestHandler(InternationalRequestHandler):
             for m in self.imported[self.host].keys():
                 sys.modules[m] = self.imported[self.host][m][0]
 
+        # for security reasons, hide some modules
+        # for virtual hosts, prevents scripts to access session information
+        save_modules = {}
+        for m in ['k_session','KarrigellRequestHandler']:
+            save_modules[m] = sys.modules[m]
+            del sys.modules[m]
+
         # self.execution is used in Include()
         self.execution=Template.ExecContext(script,self.nameSpace,
             self.path, self)
 
-        # try to reload the imported modules
-        try:
-            self.reload_modules()
-        except ReloadError:
-            return
+        # delete the imported modules whose source code has changed
+        self.reload_modules()
 
         try:
             self.outputStream.write(self.execution())
         # catches defined exceptions
-        except HTTP_REDIRECTION,url:    # HTTP redirection towards a given URL
+        except k_script.HTTP_REDIRECTION,url:    
+            # HTTP redirection towards a given URL
             self.redirect(url)
-        except HTTP_ERROR,error:
-            self.translateClassAttributes()
+        except k_script.HTTP_ERROR,error:
             self.send_error(error.code,error.message)
-        except AUTH_ABORT:
+        except k_script.HTTP_RESPONSE,(code,message):
+            self.karrigellSendResponse(code,message)
+        except k_script.AUTH_ABORT:
             pass
         else:
             # if everything's ok, send response
             self.karrigellSendResponse(200,"Ok")
 
+        # restore saved modules
+        for m in save_modules.keys():
+            sys.modules[m] = save_modules[m]
+        
         # save session object
         k_session.store(self.sessionId,self.sessionObject)
         
@@ -473,8 +612,12 @@ class KarrigellRequestHandler(InternationalRequestHandler):
                 if pyfile.endswith('.pyc'):
                     pyfile = pyfile[:-1]
                 if pyfile.endswith('.py'):
-                    self.imported[self.host][m] = [sys.modules[m],
-                        pyfile,os.stat(pyfile)[8]]
+                    try:
+                        self.imported[self.host][m] = [sys.modules[m],
+                            pyfile,os.stat(pyfile)[8]]
+                    except OSError:
+                        print 'problem with module %s' %m
+                        print 'file %s' %pyfile
 
         for m in newModules:
             # may seem strange, but *must* set the module to None first
@@ -483,9 +626,9 @@ class KarrigellRequestHandler(InternationalRequestHandler):
         sys.path = copy.copy(initial_path)
 
     def reload_modules(self):
-        # if debug mode, force reload of the imported modules
-        # whose source code has changed since last request
-        if k_config.debug:
+        # if k_config.reload_modules is set, reload the imported 
+        # modules whose source code has changed since last request
+        if k_config.reload_modules:
             to_reload = []
             deleted = []
             for m in sys.modules:
@@ -508,8 +651,11 @@ class KarrigellRequestHandler(InternationalRequestHandler):
                 except ImportError:
                     pass
                 except:
+                    # remove from sys.modules for next time
+                    del sys.modules[module]
+                    del self.imported[self.host][module]
                     # print traceback
-                    self.outputStream = cStringIO.StringIO()
+                    self.outputStream = StringIO()
                     self.outputStream.write('<font face="verdana" color="red">')
                     self.outputStream.write('<b>Error reloading module %s</b>'
                         % module)
@@ -549,11 +695,11 @@ class KarrigellRequestHandler(InternationalRequestHandler):
     def authenticate(self,realm,errorMessage):
         self.send_response(401,"Authorization")
         self.send_header("WWW-Authenticate",'Basic realm="%s"' %realm)
-        self.send_header("Content-type","text/html")
+        self.send_header("Content-type",content_type())
         self.end_headers()
-        self.wfile.write(errorMessage)
+        self.wfile.write(k_encoding.encode(errorMessage, k_config.outputEncoding))
         # message if user cancels authentication request
-        raise AUTH_ABORT
+        raise k_script.AUTH_ABORT
 
     def setSessionObject(self):
         """Internal method, initializes the session object
@@ -595,38 +741,26 @@ class KarrigellRequestHandler(InternationalRequestHandler):
         explain = long
         self.log_error("code %d, message %s", code, message)
         self.send_response(code, message)
-        self.send_header("Content-Type", "text/html")
+        self.send_header("Content-Type", content_type())
         self.end_headers()
-        # translation
+        # translation 
         gettext._translations={}
-        try:
-            t=gettext.translation("messages",
-                os.path.join(URLResolution.serverDir,"translations"),
-                self.accepted_languages)
-            t.install()
-        except:
-            t=gettext.NullTranslations()
-        self.wfile.write(self.error_message_format %
+        self.get_language()
+        try: 
+            self.t=gettext.translation("messages", 
+            os.path.join(k_config.serverDir,"translations"), 
+            self.accepted_languages) 
+        except: 
+            self.t=gettext.NullTranslations() 
+        self.t.install(unicode=1) 
+        self.translateClassAttributes()
+        err_msg = (self.error_message_format %
                          {'code': code,
-                          'message': _(message),
-                          'explain': _(explain),
+                          'message': message,
+                          'explain': explain,
                           'version': __version__,
                           'time':self.date_time_string()})
-
-    def date_time_string(self):
-        """Overrides BaseHTTPServer.BaseHTTPRequestHandler's date_time_string
-        with localtime instead of GMT"""
-        now = time.time()
-        year, month, day, hh, mm, ss, wd, y, z = time.localtime(now)
-        s = "%s, %02d %3s %4d %02d:%02d:%02d" % (
-                _(self.weekdayname[wd]),
-                day, self.monthname[month], year,
-                hh, mm, ss)
-        return s
-
-    def log_message(self, format, *args):
-        if not k_config.silent:
-            BaseHTTPServer.BaseHTTPRequestHandler.log_message(self, format, *args)
+        self.wfile.write(k_encoding.encode(err_msg, k_config.outputEncoding))
 
     def testGzip(self):
         """Test if content should be gzipped"""
@@ -635,6 +769,7 @@ class KarrigellRequestHandler(InternationalRequestHandler):
         if not gzip_support:
             return False
         accept_encoding = self.HEADERS.get('accept-encoding','').split(',')
+        accept_encoding = [ x.strip() for x in accept_encoding ]
         # if gzip is supported by the user agent,
         # and if the option gzip in the [Server] section of the
         # configuration file is set, 
@@ -669,7 +804,7 @@ class KarrigellRequestHandler(InternationalRequestHandler):
             for morsel in self.SET_COOKIE.values():
                 self.send_header('Set-Cookie', morsel.output(header='').lstrip())
         # test if content should be gzipped
-        if self.testGzip():
+        if code == 200 and self.testGzip():
             self.outputStream = self.doGzip(self.outputStream.getvalue())
         # set Content-Length header
         self.RESPONSE['Content-Length']=self.outputStream.tell()
@@ -681,9 +816,10 @@ class KarrigellRequestHandler(InternationalRequestHandler):
         self.outputStream.seek(0)
         self.copyfile(self.outputStream, self.wfile)
 
+
 # for internationalization
 t=gettext.NullTranslations()
-t.install()
+t.install(unicode=1) # unicode=1 is not needed here...
 
 # Python may not have gzip support
 gzip_support = False
@@ -696,4 +832,4 @@ except ImportError:
 
 initial_path = copy.copy(sys.path)
 initialModules = copy.copy(sys.modules.keys())
-initialModules.sort()
\ No newline at end of file
+initialModules.sort()
