@@ -38,7 +38,7 @@ s.send('GET /index.html HTTP/1.1\r\n')
 s.send('\r\n')
 res = Response(s) 
 
-print res.body == file(os.path.normpath('../index.html'),'rb').read()
+print res.body == file(os.path.normpath('../webapps/index.html'),'rb').read()
 
 s = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
 s.connect(('localhost',80))
